from importlib import import_module

from scrapy import signals
from scrapy.exceptions import NotConfigured
from scrapy.http import HtmlResponse
from selenium.webdriver.support.ui import WebDriverWait


class SeleniumDownloaderMiddleware:
    """Scrapy middleware handling the requests using selenium"""

    def __init__(self, driver_name, driver_executable_path, driver_arguments,
        browser_executable_path):
        """Initialize the selenium webdriver
        Parameters
        ----------
        driver_name: str
            The selenium ``WebDriver`` to use
        driver_executable_path: str
            The path of the executable binary of the driver
        driver_arguments: list
            A list of arguments to initialize the driver
        browser_executable_path: str
            The path of the executable binary of the browser
        """

        webdriver_base_path = f'selenium.webdriver.{driver_name}'

        driver_klass_module = import_module(f'{webdriver_base_path}.webdriver')
        driver_klass = getattr(driver_klass_module, 'WebDriver')

        driver_options_module = import_module(f'{webdriver_base_path}.options')
        driver_options_klass = getattr(driver_options_module, 'Options')

        driver_options = driver_options_klass()
        if browser_executable_path:
            driver_options.binary_location = browser_executable_path
        for argument in driver_arguments:
            driver_options.add_argument(argument)

        driver_kwargs = {
            'executable_path': driver_executable_path,
            f'{driver_name}_options': driver_options
        }

        self.driver = driver_klass(**driver_kwargs)

    @classmethod
    def from_crawler(cls, crawler):
        """Initialize the middleware with the crawler settings"""

        driver_name = crawler.settings.get('SELENIUM_DRIVER_NAME')
        driver_executable_path = crawler.settings.get('SELENIUM_DRIVER_EXECUTABLE_PATH')
        browser_executable_path = crawler.settings.get('SELENIUM_BROWSER_EXECUTABLE_PATH')
        driver_arguments = crawler.settings.get('SELENIUM_DRIVER_ARGUMENTS')

        if not driver_name or not driver_executable_path:
            raise NotConfigured(
                'SELENIUM_DRIVER_NAME and SELENIUM_DRIVER_EXECUTABLE_PATH must be set'
            )

        middleware = cls(
            driver_name=driver_name,
            driver_executable_path=driver_executable_path,
            driver_arguments=driver_arguments,
            browser_executable_path=browser_executable_path
        )

        crawler.signals.connect(middleware.spider_closed, signals.spider_closed)

        return middleware

    def process_request(self, request, spider):
        """Process a request using the selenium driver if applicable"""
        if not isinstance(request, SeleniumRequest):
            return None

        self.driver.get(request.url)

        for cookie_name, cookie_value in request.cookies.items():
            self.driver.add_cookie(
                {
                    'name': cookie_name,
                    'value': cookie_value
                }
            )

        if request.wait_until:
            WebDriverWait(self.driver, request.wait_time).until(
                request.wait_until
            )

        if request.screenshot:
            request.meta['screenshot'] = self.driver.get_screenshot_as_file('/Volumes/code/weidian/screenshot/%s.png' % request.url)

        if request.script:
            self.driver.execute_script(request.script)
        
        if request.driver_callback:
            request.driver_callback(self.driver, request)

        body = str.encode(self.driver.page_source)

        # Expose the driver via the "meta" attribute
        request.meta.update({'driver': self.driver})

        return HtmlResponse(
            self.driver.current_url,
            body=body,
            encoding='utf-8',
            request=request
        )

    def spider_closed(self):
        """Shutdown the driver when spider is closed"""

        self.driver.quit()



from scrapy import Request
class SeleniumRequest(Request):
    """Scrapy ``Request`` subclass providing additional arguments"""

    def __init__(self, wait_time=10, wait_until=None, screenshot=False, script=None, driver_callback=None, *args, **kwargs):
        """Initialize a new selenium request
        Parameters
        ----------
        wait_time: int
            The number of seconds to wait.
        wait_until: method
            One of the "selenium.webdriver.support.expected_conditions". The response
            will be returned until the given condition is fulfilled.
        screenshot: bool
            If True, a screenshot of the page will be taken and the data of the screenshot
            will be returned in the response "meta" attribute.
        script: str
            JavaScript code to execute.
        """

        self.wait_time = wait_time
        self.wait_until = wait_until
        self.screenshot = screenshot
        self.script = script
        self.driver_callback = driver_callback

        super().__init__(*args, **kwargs)


from scrapy import signals
from scrapy.http import HtmlResponse
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
import pyotp

class GoogleLoginDownloaderMiddleware(object):
    cookies = None
    login_url = 'https://accounts.google.com/signin/v2/identifier?flowName=GlifWebSignIn&flowEntry=ServiceLogin'
    # Not all methods need to be defined. If a method is not defined,
    # scrapy acts as if the downloader middleware does not modify the
    # passed objects.

    @classmethod
    def from_crawler(cls, crawler):
        # This method is used by Scrapy to create your spiders.
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def process_request(self, request, spider):
        return None

    def process_response(self, request, response, spider):
        driver = request.meta.get('driver')
        if self.cookies == None:

            driver.get(self.login_url)
            WebDriverWait(driver, 10).until(
                EC.visibility_of_element_located((By.ID, 'headingText'))
            )

            print('start login...')
            driver.find_element(By.ID, 'identifierId').send_keys(spider.settings.get('GOOGLE_ACCOUNT'))
            driver.find_element(By.CSS_SELECTOR, '#identifierNext > div > button').click()

            print('account next...')
            WebDriverWait(driver, 10).until(
                EC.visibility_of_element_located((By.CSS_SELECTOR, "#passwordNext > div > button"))
            )

            driver.find_element(By.CSS_SELECTOR, '#password > div.aCsJod.oJeWuf > div > div.Xb9hP > input').send_keys(spider.settings.get('GOOGLE_PWD'))
            driver.execute_script('document.querySelector("#passwordNext > div > button").click()')
            print('pwd next...')

            try:
                if(spider.settings.get('GOOGLE_OTPKEY')):
                    WebDriverWait(driver, 10).until(
                        EC.visibility_of_element_located((By.ID, "totpPin"))
                    )
                    driver.implicitly_wait(1)
                    totp = pyotp.TOTP(spider.settings.get('GOOGLE_OTPKEY'))
                    print('totp', totp.now())
                    driver.find_element(By.ID, 'totpPin').send_keys(totp.now())
                    driver.execute_script('document.querySelector("#totpNext > div > button").click()')
            except:
                pass

            WebDriverWait(driver, 10).until(
                EC.presence_of_element_located((By.CSS_SELECTOR, "div.yLceid"))
            )

            WebDriverWait(driver, 20).until(
                element_has_content(driver.find_element(By.TAG_NAME, 'h1').get_attribute('innerHTML'), spider.settings.get('GOOGLE_NAME'))
            )

            print('login success...')

            print('record cookies...')

            request.cookies = self.cookies = driver.get_cookies()

        driver.get(request.url)
        if (hasattr(spider, 'iframe_selector')):
            WebDriverWait(driver, 10).until(
                EC.visibility_of_element_located((By.CSS_SELECTOR, getattr(spider, 'iframe_selector')))
            )

            iframe = driver.find_element(By.CSS_SELECTOR, getattr(spider, 'iframe_selector'))
            driver.switch_to.frame(iframe)

        WebDriverWait(driver, 20).until(
            EC.visibility_of_element_located((By.CSS_SELECTOR, spider.css_selector))
        )

        body = str.encode(driver.find_element(By.TAG_NAME, 'html').get_attribute('outerHTML'))


        # Expose the driver via the "meta" attribute
        request.meta.update({'driver': driver})
        return HtmlResponse(
            request.url,
            body=body,
            encoding='utf-8',
            request=request
        )

    def process_exception(self, request, exception, spider):
        # Called when a download handler or a process_request()
        # (from other downloader middleware) raises an exception.

        # Must either:
        # - return None: continue processing this exception
        # - return a Response object: stops process_exception() chain
        # - return a Request object: stops process_exception() chain
        pass

    def spider_opened(self, spider):
        spider.logger.info('Spider opened: %s' % spider.name)

class element_has_content(object):
    def __init__(self, locator, content):
        self.locator = locator
        self.content = content

    def __call__(self, driver):
        result = self.locator.find(self.content)
        if (result != -1):
            return True
        else:
            return False
